﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;

namespace 言語バー空気嫁 {
    public partial class Form1:Form {
        IniFile settingFile = new IniFile(Path.Combine(Application.StartupPath, "settings.ini"));
        List<string> edits = new List<string>();
        List<string> notEdits = new List<string>();
        int moveCaletWait;
        int moveOutOfWindowWait;
        int ofsetX, ofsetY;
        IntPtr focusWindow = IntPtr.Zero;
        IntPtr activeWindow = IntPtr.Zero;
        Win32.RECT rect = new Win32.RECT();
        LangBer langBer = new LangBer();
        Timer moveOutOfWindowWaitTimer = null;

        public Form1() {
            InitializeComponent();
            this.notifyIcon1.Icon = this.Icon;
            this.notifyIcon1.Text = this.Text;
        }

        private void Form1_Load(object sender, EventArgs e) {
            readSettings();

            foreach(string line in File.ReadAllLines(Path.Combine(Application.StartupPath, "edit.txt"))) {
                if(line.Length == 0 || line.StartsWith("#")) {
                    continue;
                }
                edits.Add(line.Split(new char[] { '#' })[0].ToLower().Trim());
            }

            foreach(string line in File.ReadAllLines(Path.Combine(Application.StartupPath, "notedit.txt"))){
                if(line.Length == 0 || line.StartsWith("#")) {
                    continue;
                }
                notEdits.Add(line.Split(new char[] { '#' })[0].ToLower().Trim());
            }

            this.timer1.Interval = this.moveCaletWait;
        }

        private void readSettings() {
            try {
                this.moveCaletWait = int.Parse(settingFile["Wait", "MoveCaletWait"]);
                this.moveOutOfWindowWait = int.Parse(settingFile["Wait", "MoveOutOfWindowWait"]);
                this.ofsetX = int.Parse(settingFile["Ofset", "X"]);
                this.ofsetY = int.Parse(settingFile["Ofset", "Y"]);
            } catch {
                MessageBox.Show("settings.iniの内容が不正です。アプリケーションを終了します。");
                this.Close();
            }
        }

        private void Form1_Shown(object sender, EventArgs e) {
            this.Hide();
        }

        private void timer1_Tick(object sender, EventArgs e) {
            Win32.GUITHREADINFO guiThreadInfo = this.getGUITHREADINFO();
            Win32.RECT tmpRECT = this.getWindowRECT(guiThreadInfo.hwndActive);

            if(this.activeWindow != guiThreadInfo.hwndActive) {
                Timer timer = new Timer();
                timer.Interval = this.moveCaletWait;
                timer.Tick += (s_, e_) => {
                    this.moveLangBar();
                    timer.Dispose();
                };
                timer.Start();
            } else if(this.focusWindow != guiThreadInfo.hwndFocus) {
                this.moveLangBar(guiThreadInfo);
            } else if(this.rect != tmpRECT) {
                this.moveLangBarToOutOfWindow(guiThreadInfo);
            }

            this.focusWindow = guiThreadInfo.hwndFocus;
            this.activeWindow = guiThreadInfo.hwndActive;
            this.rect = tmpRECT;
        }

        private Win32.RECT getWindowRECT(IntPtr handle) {
            Win32.RECT rect = new Win32.RECT();
            Win32.GetWindowRect(handle, out rect);
            return rect;
        }

        private Win32.GUITHREADINFO getGUITHREADINFO() {
            uint dummy = 0;
            Win32.GUITHREADINFO guiThreadInfo = new Win32.GUITHREADINFO();
            guiThreadInfo.cbSize = (uint)Marshal.SizeOf(guiThreadInfo);
            uint threadID = Win32.GetWindowThreadProcessId(Win32.GetForegroundWindow(), ref dummy);
            Win32.GetGUIThreadInfo(threadID, ref guiThreadInfo);
            return guiThreadInfo;
        }

        private void moveLangBar() {
            this.moveLangBar(this.getGUITHREADINFO());
        }

        private void moveLangBar(Win32.GUITHREADINFO guiThreadInfo) {
            if(this.isEdit(guiThreadInfo.hwndFocus)) {
                this.moveLangBarToCalet(guiThreadInfo);
                if(this.moveOutOfWindowWaitTimer != null) {
                    this.moveOutOfWindowWaitTimer.Dispose();
                }
                this.moveOutOfWindowWaitTimer = new Timer();
                this.moveOutOfWindowWaitTimer.Interval = this.moveOutOfWindowWait;
                this.moveOutOfWindowWaitTimer.Tick += (sender, e) => {
                    this.moveLangBarToOutOfWindow(guiThreadInfo);
                    this.moveOutOfWindowWaitTimer.Dispose();
                };
                this.moveOutOfWindowWaitTimer.Start();
            } else {
                this.moveLangBarToOutOfWindow(guiThreadInfo);
            }
        }

        private void moveLangBarToCalet(Win32.GUITHREADINFO guiThreadInfo) {
            Win32.POINT p = new Win32.POINT();
            p.x = guiThreadInfo.rcCaret.right + this.ofsetX;
            p.y = guiThreadInfo.rcCaret.bottom + this.ofsetY;

            Win32.ClientToScreen(guiThreadInfo.hwndFocus, ref p);

            langBer.Location = p;
        }

        private void moveLangBarToOutOfWindow(Win32.GUITHREADINFO guiThreadInfo) {
            Win32.POINT p = new Win32.POINT();
            Win32.RECT r;
            Win32.GetWindowRect(guiThreadInfo.hwndActive, out r);
            p.x = r.left;
            p.y = r.top - (langBer.Bounds.bottom - langBer.Bounds.top);

            langBer.Location = p;
        }

        private bool isEdit(IntPtr handle) {
            bool isEdit = false;
            StringBuilder className = new StringBuilder(256);
            Win32.GetClassName(handle, className, className.Capacity);
            string classNameToLower = className.ToString().ToLower();

            foreach(string editClass in this.edits) {
                if(classNameToLower.Contains(editClass)) {
                    isEdit = true;
                    break;
                }
            }

            foreach(string notEditClass in notEdits) {
                if(classNameToLower.Contains(notEditClass)) {
                    isEdit = false;
                    break;
                }
            }

            if(this.textBox1.Handle != handle) {
                this.textBox1.Text = className.ToString() + Environment.NewLine +
                  "IsEdit=" + isEdit.ToString();
            }

            return isEdit;
        }

        private void 終了XToolStripMenuItem_Click(object sender, EventArgs e) {
            this.FormClosing -= Form1_FormClosing;
            this.Close();
        }
         
        private void Form1_FormClosing(object sender, FormClosingEventArgs e) {
            if(e.CloseReason != CloseReason.TaskManagerClosing && e.CloseReason != CloseReason.WindowsShutDown) {
                this.Hide();
                e.Cancel = true;
            }
        }

        private void notifyIcon1_MouseClick(object sender, MouseEventArgs e) {
            if(e.Button == MouseButtons.Left) {
                this.Visible = !this.Visible;
            }
        }
    }
}